class IOCallbacksStorage(object):
    """ This class stores user-defined callbacks. Feel free to inherit this! """
    def __init__(self, env):
        """
        The base interface for inputs and outputs for a asciidots program.

        :type env: dots.environment.Env
        """
        self.env = env
        self.env.io = self

    def get_input(self):
        """ This function should return an integer from stdin, the user, etc """
        raise Exception('get_input: not implemented')

    def on_output(self, value):
        """
        This function is called whenever the AsciiDots program returns a value
        Note that the value may be a string
        """
        raise Exception('on_output: not implemented')

    def on_finish(self):
        """ This function is called when the AsciiDots program finished executing """
        raise Exception('on_finish: not implemented')

    def on_error(self, error_text):
        """ This function is pretty self explanatory; it catches errors with the AsciiDot program that occur during interpretetion """
        raise Exception('on_error: not implemented')

    def on_microtick(self, dot):
        pass


class IOCallbacksStorageConstructor(object):
    """ This is just a way to build an IOCallbacksStorage equivalent object """
    def __init__(self, get_input, on_output, on_finish, on_error, on_microtick):
        self.get_input = get_input
        self.on_output = on_output
        self.on_finish = on_finish
        self.on_error = on_error
        self.on_microtick = on_microtick



"""
Here are defined the constants of the language.
You will alos find shortcuts for repetitive values in the code like UP and DOWN.
"""

from dots.vector import Pos

# Defining directions
UP = Pos(0, -1)
DOWN = Pos(0, 1)
LEFT = Pos(-1, 0)
RIGHT = Pos(1, 0)

DIRECTIONS = (UP, RIGHT, DOWN, LEFT)


# Defining Chars
COMMENT_START = '``'
ID_MODE = '@'
VALUE = '#'



#!/usr/bin/env python
# coding=utf-8

import codecs
import locale
import os
import signal
import sys
import time

import click

from dots.environment import Env
from dots.exceptions import DotsExit

if codecs.lookup(locale.getpreferredencoding()).name == 'ascii':
    os.environ['LANG'] = 'en_US.utf-8'

from dots.interpreter import AsciiDotsInterpreter
from dots.callbacks import IOCallbacksStorage

from dots import terminalsize

try:
    import curses
except ImportError:
    print('failed to import curses; running in compatibility mode')
    compat_debug_default = True
else:
    compat_debug_default = False

terminal_lines = terminalsize.get_terminal_size()[1]
default_debug_lines = int(terminal_lines * 2 / 3)

interpreter = None

debug_ = True
autostep_debug_ = False


class DefaultIOCallbacks(IOCallbacksStorage):
    """The default class to manage the input and output of a dots program."""

    def __init__(self, env, ticks, silent, debug, compat_debug, debug_lines, autostep_debug, output_limit):
        """

        :param dots.environment.Env env: The env of the interpreter
        :param int ticks: The max number of ticks for the program
        :param bool silent: True to turn off all outputs
        :param bool debug: True to show the execution of the program
        :param bool compat_debug: True to show the debug with only builtin functions
        :param int debug_lines: The number of lines to show the debug
        :param float autostep_debug: The timebetween automatic ticks. 0 disables the auto ticks.
        :param int output_limit: The max number of outputs for the program
        """
        super().__init__(env)

        # if it is zero or false, we don't want to stop
        self.ticks_left = ticks or float('inf')
        self.outputs_left = output_limit or float('inf')

        self.silent = silent
        self.debug = debug
        self.compat_debug = compat_debug
        self.debug_lines = debug_lines
        self.debug_cols = terminalsize.get_terminal_size()[0] - 1
        self.autostep_debug = autostep_debug

        self.compat_logging_buffer = ''
        self.compat_logging_buffer_lines = terminal_lines - debug_lines - 1

        self.first_tick = True

        if self.debug and not self.compat_debug:
            self.logging_loc = 0
            self.logging_x = 1

            self.stdscr = curses.initscr()

            curses.start_color()

            curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)
            curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
            curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)
            curses.init_pair(4, curses.COLOR_BLUE, curses.COLOR_BLACK)

            curses.noecho()

            # hides the cursor
            curses.curs_set(False)

            # defining the two main parts of the screen: the view of the program
            self.win_program = curses.newwin(self.debug_lines, curses.COLS, 0, 0)
            # and pad for the output of the prog
            self.logging_pad = curses.newpad(1000, curses.COLS - 1)

            def signal_handler(signal, frame):
                self.on_finish()
                sys.exit(0)

            signal.signal(signal.SIGINT, signal_handler)

    def get_input(self):
        """Get an input from the user."""
        if not self.debug or self.compat_debug:
            if not sys.stdin.isatty():
                return input('')
            else:
                return input('?: ')
        else:
            return self.curses_input(self.stdscr, curses.LINES - 3, 2, '?: ')

    def curses_input(self, stdscr, row, col, prompt_string):
        """
        Get an input string with curses.

        Row and col are the start position ot the prompt_string.
        """
        curses.echo()
        stdscr.addstr(row, col, str(prompt_string), curses.A_REVERSE)
        stdscr.addstr(row + 1, col, " " * (curses.COLS - 1))
        stdscr.refresh()
        input_val = ""

        while len(input_val) <= 0:
            input_val = stdscr.getstr(row + 1, col, 20)

        return input_val

    def on_output(self, value):
        value = str(value)
        self.outputs_left -= 1

        # maximum output reached, we quit the prog
        if self.outputs_left == 0:
            raise DotsExit

        # no printing mode
        if self.silent:
            return

        if not self.debug:
            print(value, end='', flush=True)

        elif self.compat_debug:
            # we add the ouput to the buffer
            self.compat_logging_buffer += value
            # and we keep the maximum number of line to compat_logging_buffer_lines
            self.compat_logging_buffer = '\n'.join(
                self.compat_logging_buffer.split('\n')[-self.compat_logging_buffer_lines:])

        else:
            # add the output string to the pad
            self.logging_pad.addstr(self.logging_loc, self.logging_x, str(value))
            self.logging_pad.refresh(self.logging_loc - min(self.logging_loc, curses.LINES -
                                                            self.debug_lines - 1),
                                     0, self.debug_lines, 0, curses.LINES - 1, curses.COLS - 1)

            # FIXME: This should count the number of newlines instead
            if str(value).endswith('\n'):
                self.logging_loc += 1
                self.logging_x = 1
            else:
                self.logging_x += len(value)

    def on_finish(self):
        """Close cleanly the I/O."""
        # we need to close curses only if it was opened
        if self.debug and not self.compat_debug:
            curses.nocbreak()
            self.stdscr.keypad(False)
            curses.echo()

            curses.endwin()

    def on_error(self, error_text):
        """Show the error and cleans the I/O."""
        self.on_output('error: {}'.format(error_text))

        raise DotsExit

    def on_microtick(self, dot):

        # we want to show the program
        if self.debug and not self.silent:

            display_y = 0
            last_line_is_empty = False
            dots_position_list = [d.pos for d in self.env.dots if not d.state.is_dead()]

            # cleaning the screen
            if self.compat_debug:
                if os.name == 'nt':
                    os.system('cls')  # For Windows
                else:
                    os.system('clear')  # For Linux/OS X
            else:
                # Sync the screen with the previous changes
                self.win_program.refresh()

            for y, line in enumerate(self.env.world.map):
                # don't show more lines than the debug view has
                if display_y > self.debug_lines - 2:
                    break

                # if the line is empty
                # The aim of this is to skip blocks of white lines (ie. if there is a lot of comments)
                if not ''.join(line).rstrip():
                    if last_line_is_empty:
                        continue
                    else:
                        last_line_is_empty = True
                else:
                    last_line_is_empty = False

                for x, char in enumerate(line):
                    # curses crash when you try to print outside the screen so don't
                    if x > self.debug_cols - 1:
                        break

                    if char == '\n':
                        # The new line in printed only in compat mode, at the end of the loop
                        continue

                    # Printing each char with the right color
                    if (x, y) in dots_position_list:
                        # We want to always see the dot, so if is on a space (printing) we render a .
                        if char == ' ':
                            self.print_char('.', 1, display_y, x)
                        else:
                            self.print_char(char, 1, display_y, x)
                    elif char.isLibWarp():
                        self.print_char(char, 2, display_y, x)
                    elif char.isWarp():
                        self.print_char(char, 3, display_y, x)
                    elif char in '#@~' or char.isOper():
                        self.print_char(char, 4, display_y, x)
                    else:
                        self.print_char(char, 0, display_y, x)

                if self.compat_debug:
                    print()

                # The line pos. NB: This is different to y because we can skip lines)
                display_y += 1

            # print the output part
            if self.compat_debug:
                print('\n' + self.compat_logging_buffer, end='', flush=True)

            if not self.first_tick or True:
                # step automatically or wait for input
                if self.autostep_debug:
                    time.sleep(self.autostep_debug)
                else:
                    if self.compat_debug:
                        input("Press enter to step...")
                    else:
                        # we need to get the key pressed and check if it is ^C to quit, because
                        # curses will intercept those inputs who wont go in the python interpreter
                        # Who will stop the execution
                        keycode = self.stdscr.getch()

                        if keycode in (3, 26):
                            raise DotsExit
            else:
                self.first_tick = False

        self.ticks_left -= 1
        if self.ticks_left == 0:
            self.on_output('QUITTING next step!\n')

            raise DotsExit

    def print_char(self, char, color_code, row=None, col=None):
        """
        Print one char to the screen with coloration.

        In compat_debug this will just append the char to stdout. This checks for silent mode
        :param char: The character to print
        :param color_code: The colorcode 1234 = RGYB
        :param row: The y pos used with curses
        :param col: The x pos used with curses
        """

        if self.silent:
            return 42

        if self.compat_debug:
            if not color_code:
                # Zero is the regular print, but the code 33 will draw black over black and we dont want that
                print(char, end='')
            else:
                print('\033[0;3', color_code, 'm', char, '\033[0m', sep='', end='')
        else:
            self.win_program.addstr(row, col, char, curses.color_pair(color_code))


@click.command()
@click.argument('filename')
@click.option('--debug', '-d', is_flag=True, help='Show the execution of the program and the course of the dots.')
@click.option('--autostep_debug', '-a', default=False, help='The time between every tick')
@click.option('--output_limit', '-o', default=0, help='Terminate the program after N outputs.')
@click.option('--ticks', '-t', default=0, help='Terminate the program after N ticks.')
@click.option('--silent', '-s', is_flag=True, help='No printing, for benchmarking.')
@click.option('--compat_debug', '-w', is_flag=True, help='Force the debug rendering without ncurses.')
@click.option('--debug_lines', '-l', default=default_debug_lines, help='The size of the debug view.')
@click.option('--async', '-y', is_flag=True, help='Only one dot moves at a time. Easier to debug.')
def main(filename, ticks, silent, debug, compat_debug, debug_lines, autostep_debug, output_limit, async):
    global interpreter

    if autostep_debug is not False:
        autostep_debug = float(autostep_debug)

    compat_debug = compat_debug or compat_debug_default

    run_in_parallel = not async

    env = Env()
    env.io = DefaultIOCallbacks(env, ticks, silent, debug, compat_debug, debug_lines, autostep_debug, output_limit)

    program_dir = os.path.dirname(os.path.abspath(filename))
    with open(filename, encoding='utf-8') as file:
        program = file.read()

    interpreter = AsciiDotsInterpreter(env, program, program_dir, run_in_parallel)
    interpreter.run()


if __name__ == "__main__":
    main()



class Pos(tuple):
    def __new__(cls, x, y=None):
        if y is None:
            return tuple.__new__(cls, x)
        return tuple.__new__(cls, (x, y))

    def __str__(self):
        return "<Pos({}, {})>".format(*self)

    def __add__(self, other):
        return Pos(self[0] + other[0], self[1] + other[1])

    __radd__ = __add__

    def __neg__(self):
        return Pos(-self[0], -self[1])

    def __sub__(self, other):
        return Pos(self[0] - other[0], self[1] - other[1])

    def __rsub__(self, other):
        return Pos(other[0] - self[0], other[1] - self[1])

    def __hash__(self):
        return super().__hash__()

    @property
    def x(self):
        return self[0]

    @property
    def y(self):
        return self[1]

    @property
    def row(self):
        return self[1]

    @property
    def col(self):
        return self[0]



# Char inheritance works similar to the Decorator Pattern

# TODO: make the iters have a default arg of self._data_array?

import os

from dots.vector import Pos
from .chars import *


class World(object):
    def __init__(self, env, world_map, program_dir):
        """
        Create a new world to do dots races !

        :param dots.environment.Env env: The environment for the program
        :param str world_map: The string representing the world.
        :param str program_dir: The directory of the program
        """

        self.env = env
        self.env.world = self

        self.program_dir = program_dir

        self.map = self.map_from_raw(world_map)

        self._worldwide_warp_id_counter = 0
        self._setup_warps_for(self.map)

        self._import_libraries()

        self._setup_operators()

        self._connect_warps()

        self._update_class_of_dots()

    def get_coords_of_dots(self):
        """Yiels the cordinates of every dot char in the world."""
        for y, line in enumerate(self.map):
            if line[0] == '%':
                continue

            for x, char in enumerate(line):
                if char.isDot():
                    yield Pos(x, y)

    # ✓
    def get_char_at(self, pos: Pos):
        """Get the Char at the given position."""
        # NOTE: _data_array has to be accesed using y, x due to the way it is created
        return self.map[pos.row][pos.col]

    # ✓
    def does_loc_exist(self, loc: Pos):
        """True if this location exists on the map."""
        return 0 <= loc.row < len(self.map) and 0 <= loc.col < len(self.map[loc.row])

    def is_char_at(self, pos, char):
        """True iff the pos exists and is the char is the same."""
        if not self.does_loc_exist(pos):
            return False
        return self.get_char_at(pos) == char

    # NOTE: Hopefully done?
    def _import_libraries(self, map=None):
        """
        Import the library for a given map.

        :param str map: The map to import libraries from. Defaults to the world map.
        """

        if map is None:
            map = self.map

        lib_filenames_for_chars = self._get_lib_files_by_alias(map)
        lib_chars = lib_filenames_for_chars.keys()

        self._update_class_of_lib_chars(map, lib_chars)

        singleton_ids = {}

        for y, line in enumerate(map):
            if line[0] == '%':
                continue

            for x, char in enumerate(line):
                if char.isLibWarp() and char.get_id() is None:
                    if char not in singleton_ids:
                        this_warp_id = self._worldwide_warp_id_counter

                        map[y][x].set_id(this_warp_id)
                        self._worldwide_warp_id_counter += 1

                        if char.isSingletonLibWarp():
                            singleton_ids[char] = this_warp_id

                        if char in lib_filenames_for_chars:
                            filename = lib_filenames_for_chars[char]

                            self._import_lib_file_with_warp_id(map, filename, this_warp_id,
                                                               is_singleton=char.isSingletonLibWarp())
                    else:
                        map[y][x].set_id(singleton_ids[char])

    # NOTE: Hopefully done?
    def _import_lib_file_with_warp_id(self, char_obj_array, filename, warp_id, is_singleton):
        path = self._get_path_of_lib_file(filename)

        with open(path, 'r') as lib_file:
            lib_code = lib_file.read()

        lib_char_obj_array = self.map_from_raw(lib_code)

        exposed_char_str = None

        for y, char_list in enumerate(lib_char_obj_array):
            line = ''.join(char_list).rstrip()

            if line[:2] == '%+':
                print(('%+ notation has become replaced by a new notation\n' +
                       'you now define a single warp char as an entry point to your code using %$\n' +
                       'for this code, it is recommended that your replace\n\n' +
                       '%+{0}{1}{2}{3}\n\n' +
                       'with\n\n' +
                       '%^X `` make sure that X doesn\'t conflict with anything\n' +
                       '%${0}{1}{2}{3}\n\n' +
                       '  {3}\n' +
                       '  |\n' +
                       '{2}-X-{0}\n' +
                       '  |\n' +
                       '  {1}\n').format(*line[2:]))

                raise Exception('obsolete code (unable to run)')
            elif line[:2] == '%^':
                exposed_char_str = line[2]  # FIXME: This only allows exposing one char!
            elif len(line) > 0 and line[0] == '%':
                continue
            else:
                for x, char in enumerate(char_list):
                    if char == exposed_char_str:
                        if is_singleton:
                            lib_char_obj_array[y][x] = SingletonLibInnerWarpChar(char)
                        else:
                            lib_char_obj_array[y][x] = LibWarpChar(char)

                        lib_char_obj_array[y][x].set_id(warp_id)

        self._setup_warps_for(lib_char_obj_array)

        self._import_libraries(lib_char_obj_array)

        char_obj_array.extend(lib_char_obj_array)

    # ✓
    def _update_class_of_lib_chars(self, char_obj_array, lib_chars):
        is_singleton_dict = self._get_dict_of_is_singleton_for_lib_chars_in(char_obj_array)

        for y, line in enumerate(char_obj_array):
            if line[0] == '%':
                continue

            for x, char in enumerate(line):
                if char in lib_chars:
                    if is_singleton_dict[char]:
                        char_obj_array[y][x] = SingletonLibOuterWarpChar(char)
                    else:
                        char_obj_array[y][x] = LibWarpChar(char)

    # ✓
    def _get_path_of_lib_file(self, filename):
        path_for_inside_program_dir = os.path.join(self.program_dir, filename)

        # Does it exist?
        if os.path.isfile(path_for_inside_program_dir):
            return path_for_inside_program_dir
        else:  # if not, look in the libs folder
            current_dir = '/'.join(os.path.realpath(__file__).split('/')[:-1])
            path_for_inside_module = os.path.join(current_dir, 'libs', filename)

            if os.path.isfile(path_for_inside_module):
                return path_for_inside_module
            else:
                interpreter_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
                return os.path.join(interpreter_dir, "libs", filename)

    # ✓✓
    @staticmethod
    def _get_lib_files_by_alias(map):
        """
        Get the librairy files by alias char defined.

        :param List[List[Char]] map: The map to import
        """

        filename_by_alias = {}

        for row in map:
            # get back the string from the Char array and remove the trailling white spaces
            line = ''.join(row).rstrip()

            # if it's an import
            if line.startswith('%!'):
                # Retrieve the filename and the alias from the line
                pieces = line[2:].split(' ')
                filename = pieces[0]
                alias = pieces[1]

                # add it to the mapping
                filename_by_alias[alias] = filename

        return filename_by_alias

    # ✓
    def _get_dict_of_is_singleton_for_lib_chars_in(self, char_obj_array):
        is_singleton_dict = {}

        for char_list in char_obj_array:
            line = ''.join(char_list).rstrip()

            if line[:2] == '%!':
                pieces = line[2:].split(' ')

                char = pieces[1]

                is_singleton_dict[char] = True

                # if len(pieces) >= 3 and pieces[2] == '&':
                #     is_singleton_dict[char] = True
                # else:
                #     is_singleton_dict[char] = False

        return is_singleton_dict

    # ✓
    def _connect_warps(self):
        for y, line in enumerate(self.map):
            if line[0] == '%':
                continue

            for x, char in enumerate(line):
                if char.isWarp() and not isinstance(char, SingletonLibInnerWarpChar):
                    warp_id = char.get_id()
                    companion_warp_loc = self._find_companion_warp_char_loc_of(char, warp_id, Pos(x, y))

                    if companion_warp_loc is not None:
                        self.map[y][x].set_dest_loc(companion_warp_loc)

    # ✓
    def _find_companion_warp_char_loc_of(self, orig_char, warp_id, orig_pos: Pos):
        for y, line in enumerate(self.map):
            if line[0] == '%':
                continue

            for x, char in enumerate(line):
                if char.isWarp() and char.get_id() == warp_id and orig_pos != (x, y):
                    if isinstance(orig_char, SingletonLibOuterWarpChar):
                        if not isinstance(char, SingletonLibInnerWarpChar):
                            continue
                    return Pos(x, y)

    # ✓
    def _setup_warps_for(self, char_obj_array):
        self._correct_class_of_warp_chars_in(char_obj_array)

        # {letter: id}
        assigned_ids_for_letters = {}

        for (x, y), char in self._char_obj_array_iter_with_coords(char_obj_array):
            if char.isWarp() and char.get_id() is None:
                if char in assigned_ids_for_letters:
                    char_obj_array[y][x].set_id(assigned_ids_for_letters[char])
                else:
                    this_id = self._worldwide_warp_id_counter
                    assigned_ids_for_letters[char] = this_id
                    char.set_id(this_id)

                    self._worldwide_warp_id_counter += 1

    # ✓
    def _correct_class_of_warp_chars_in(self, char_obj_array):
        warp_list = self._get_warp_chars_list_from(char_obj_array)

        for y, line in enumerate(char_obj_array):
            for x, char in enumerate(line):
                if char in warp_list:
                    char_obj_array[y][x] = WarpChar(char)

    # TODO check if the char is inside of a ascii dots text string
    def _update_class_of_dots(self):
        for y, char_list in enumerate(self.map):
            last_was_backtick = False
            for x, char in enumerate(char_list):
                if char == '`':
                    if not last_was_backtick:
                        last_was_backtick = True
                    else:
                        break

                if char == '.':
                    self.map[y][x] = DotChar(char)

    # ✓
    def _setup_operators(self):
        for y, line in enumerate(self.map):
            for x, char in enumerate(line):
                if 0 < x < len(line) - 1:
                    if line[x - 1] == '{' and line[x + 1] == '}':
                        self.map[y][x] = CurlyOperChar(char)
                    elif line[x - 1] == '[' and line[x + 1] == ']':
                        self.map[y][x] = SquareOperChar(char)

    # ✓
    def _get_warp_chars_list_from(self, char_obj_array):
        warp_chars = []

        for char_list in char_obj_array:
            line = ''.join(char_list).rstrip()

            if line[:2] == '%$':
                string_with_chars = line[2:]
                string_with_chars = string_with_chars.rstrip()
                list_with_chars = list(string_with_chars)
                warp_chars.extend(list_with_chars)

        return warp_chars

    # ✓
    def _char_obj_array_iter(self, obj_array):
        for char_list in obj_array:
            for char in char_list:
                yield char

    # ✓
    def _char_obj_array_iter_with_coords(self, obj_array):
        for y, char_list in enumerate(obj_array):
            for x, char in enumerate(char_list):
                yield Pos(x, y), char

    # ✓✓
    @staticmethod
    def map_from_raw(raw_map: str):
        """
        Convert a code in a string to a usable map.

        This will suppress the comments and convert each chr of the string to the corresponding Char.
        Creates a 2D array accessible by map[row][col].
        :param str raw_map: The program as it is stored in files.
        """

        map = []

        # for each line
        for raw_line in raw_map.split('\n'):
            # removing the comments
            line = raw_line.partition('``')[0] + ' '
            # Convert the str to a list of Char
            line = [Char(c) for c in line]
            # add aech row to the map
            map.append(line)

        return map



import sys

from dots.exceptions import DotsExit
from .states import *


class Dot:
    def __init__(self, env, pos, id_=None, value=None, direction=None, state=None, stack=None):
        """
        The base unit of and ascii dot code : the dot.

        :param dots.environment.Env env: The environment for the program
        :param dots.vector.Pos pos: The position of the dot in the map
        :param float id_: the id of the dot
        :param float value: its value
        :param dots.vector.Pos direction: The direction of the dot
        :param state: Its actual state
        :param list stack:
        """

        self.pos = pos

        self.env = env
        self.id = id_ or 0
        self.value = value or 0
        self.state = state(self) if state else TravelState(self)  # type: State
        self.dir = direction or self._calculate_direction()
        self.stack = stack or []

    def __repr__(self):
        return '<Dot pos={pos}, id={id}, value={value}, dir={dir}, stack={stack}>'.format(**self.__dict__)

    def move(self):
        """Move the dot according to its direction."""
        self.pos += self.dir

    def copy(self):
        """
        Create a copy of the dot.

        :remark: The copy doesn't share any variable (by reference) to the old one. (except env of course)
        """

        return Dot(self.env, self.pos, self.id, self.value, self.dir, type(self.state), self.stack[:])

    def simulate_tick(self, run_until_waiting):
        """
        Update the dot to its next state.

        :param bool run_until_waiting: if false, the dot will perform only one tick, else it will run untill waiting
        """

        past_locations = []

        while True:

            # we need to update the screen if we keep this dot running, nobody will ever do it otherwise
            if run_until_waiting:
                self.env.io.on_microtick(self)

            # If it was already at this location, run someone else (prevent infinite loops)
            if self.pos in past_locations:
                return
            past_locations.append(self.pos)

            # If outside the map, he dies.
            if not self.env.world.does_loc_exist(self.pos):
                self.state = DeadState(self)
                return

            char = self.env.world.get_char_at(self.pos)

            # update the dot
            self.state = self.state.next(char)
            self.state.run(char)

            if self.state.is_dead():
                return

            if not self.env.world.does_loc_exist(self.pos):
                self.state = DeadState(self)
                return

            if self.env.world.is_char_at(self.pos, ' ') and not isinstance(self.state, PrintState):
                self.state = DeadState(self)
                return

            if self.state.isWaiting:
                break

            if not run_until_waiting:
                break

    def next(self):
        """Update the state of this dot accoding to the char it is in."""
        if not self.env.world.does_loc_exist(self.pos):
            self.state = DeadState(self)
            return

        char = self.env.world.get_char_at(self.pos)
        self.state = self.state.next(char)

        if char == ' ' and not isinstance(self.state, PrintState):
            self.state = DeadState(self)

    def run(self):
        """Perform the state action on this char."""
        if not self.env.world.does_loc_exist(self.pos):
            self.state = DeadState(self)
            return

        self.state.run(self.env.world.get_char_at(self.pos))

        if not self.env.world.does_loc_exist(self.pos):
            self.state = DeadState(self)
        if self.env.world.is_char_at(self.pos, ' ') and not isinstance(self.state, PrintState):
            self.state = DeadState(self)

    def _calculate_direction(self):
        """Calculate the inial direction of a just created dot."""
        valid_chars = r'\/*^v><+'

        for direction in DIRECTIONS:
            loc = self.pos + direction

            # we have no interest in chars outside
            if not self.env.world.does_loc_exist(loc):
                continue

            if direction in (UP, DOWN) and self.env.world.get_char_at(loc) == '|':
                return direction

            if direction in (LEFT, RIGHT) and self.env.world.get_char_at(loc) == '-':
                return direction

            if self.env.world.get_char_at(loc) in valid_chars:
                return direction

        # If we get here without returning, the dot can't find a direction to go!
        self.env.io.on_error("dot cannot determine location...\nx: {}, y: {}".format(*self.pos))

        self.state = DeadState(self)
        return Pos(0, 0)



class Char(str):
    def __init__(self, value):
        self.value = value

    def isDot(self):
        return False

    def isOper(self):
        return False

    def isCurlyOper(self):
        return False

    def isSquareOper(self):
        return False

    def isWarp(self):
        return False

    def isLibWarp(self):
        return False

    def isSingletonLibWarp(self):
        return False

    def isSingletonLibReturnWarp(self):
        return False


class DotChar(Char):
    def isDot(self):
        return True


class OperChar(Char):
    def __init__(self, value):
        super().__init__(value)

        self.func = None

    def isOper(self):
        return True

    def calc(self, x, y):
        if self.func is None:
            function_dict = {
                '+': (lambda x, y: x + y),
                '-': (lambda x, y: x - y),
                '*': (lambda x, y: x * y),
                '/': (lambda x, y: x / y),
                '^': (lambda x, y: x ** y),
                '%': (lambda x, y: x % y),

                'o': (lambda x, y: x | y),
                'x': (lambda x, y: x ^ y),
                '&': (lambda x, y: x & y),
                '!': (lambda x, y: x != y),

                '=': (lambda x, y: x == y),
                '>': (lambda x, y: x > y),
                'G': (lambda x, y: x >= y),
                '<': (lambda x, y: x < y),
                'L': (lambda x, y: x <= y)
            }

            self.func = function_dict[self]

        return self.func(x, y)


class CurlyOperChar(OperChar):
    def isCurlyOper(self):
        return True


class SquareOperChar(OperChar):
    def isSquareOper(self):
        return True


class WarpChar(Char):
    def __init__(self, value):
        super().__init__(value)

        self._teleporter_id = None
        self._dest_loc = None

    def isWarp(self):
        return True

    def set_id(self, teleporter_id):
        self._teleporter_id = teleporter_id

    def get_id(self):
        return self._teleporter_id

    def set_dest_loc(self, pos):
        self._dest_loc = pos

    def get_dest_loc(self):
        return self._dest_loc


class LibWarpChar(WarpChar):
    def isLibWarp(self):
        return True


class LibInnerWarpChar(LibWarpChar):
    def isLibWarp(self):
        return True

    def set_dest_loc(self, pos):
        raise Exception(
            "SingletonLibReturnWarpChar: cannot set destination; use the stack!")

    def get_dest_loc(self):
        raise Exception(
            "SingletonLibReturnWarpChar: unknown destination; use the stack!")


class LibOuterWarpChar(LibWarpChar):
    def isLibWarp(self):
        return True


# NB: Singleton refers to the library written in ascii dots, not the class itself!
class SingletonLibWarpChar(LibWarpChar):
    def isSingletonLibWarp(self):
        return True


# NB: Singleton refers to the library written in ascii dots, not the class itself!
class SingletonLibOuterWarpChar(SingletonLibWarpChar):
    pass


# NB: Singleton refers to the library written in ascii dots, not the class itself!
class SingletonLibInnerWarpChar(LibInnerWarpChar, SingletonLibWarpChar):
    pass



import threading

from dots.exceptions import DotsExit
from .dot import Dot
from .world import World


class AsciiDotsInterpreter(object):
    def __init__(self, env, program, program_dir, run_in_parallel):
        """
        Create a new instance of the interpreter to run the program.

        :param dots.environment.Env env: The environment for the program
        :param str program: The code of the program
        :param str program_dir: The path to the program directory
        :param bool run_in_parallel: temporarily, changes the way dots move : one by one or all at the same time
        """

        self.env = env
        self.env.interpreter = self
        self.env.world = World(env, program, program_dir)
        self.env.dots = []

        self.needs_shutdown = False

        self._setup_dots()
        self.run_in_parallel = run_in_parallel

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate()
        self.env.io.on_finish()

        if isinstance(exc_tb, DotsExit):
            # we don't want the exception re-raised
            return True

    def _setup_dots(self):
        """Fill the dot list with dots from the starting points in the world."""

        self.env.dots = []
        for pos in self.env.world.get_coords_of_dots():
            new_dot = Dot(self.env, pos)

            self.env.dots.append(new_dot)

    def run(self, run_in_separate_thread=None, make_thread_daemon=None):
        """
        Start executing the AsciiDots code

        Arguments:
        run_in_separate_thread -- If set to True, the program will be interpreted in a separate thread
        make_thread_daemon -- Controls whether a thread created by enabling in_seperate_thread will be run as daemon
        """

        if run_in_separate_thread:
            inter_thread = threading.Thread(target=self.run, daemon=make_thread_daemon)
            inter_thread.start()
            return

        with self:
            while not self.needs_shutdown and len(self.env.dots) > 0:

                if self.run_in_parallel:
                    self.parallel_tick()
                else:
                    self.async_tick()

            raise DotsExit

    def parallel_tick(self):
        """Simulate a tick in parallele mode"""

        dots = self.env.dots[:]
        for dot in dots:
            dot.next()

        for dot in dots:
            dot.run()

        for dot in dots:
            if dot.state.is_dead():
                self.env.dots.remove(dot)

        self.env.io.on_microtick(None)

    def async_tick(self):
        """Simulate a tick in async mode."""
        for dot in self.env.dots[:]:
            dot.simulate_tick(not self.run_in_parallel)

            if dot.state.is_dead():
                self.env.dots.remove(dot)

    def terminate(self):
        """The program will shut down at the next operation."""
        self.needs_shutdown = True



# based on https://github.com/pypa/sampleproject/blob/master/setup.py

import setuptools
from setuptools import setup, find_packages

from codecs import open
from os import path

import glob

try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst', 'md')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()


setup(
    name='asciidots',

    version='1.0.6',

    description='Interpreter for the AsciiDots esolang.',
    long_description=read_md('README.md'),

    url='https://github.com/aaronduino/asciidots',

    author='Aaron Janse',
    author_email='gitduino@gmail.com',

    license='APGL-v3.0',

    python_requires='>=3',

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Education',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Science/Research',
        'Topic :: Games/Entertainment',
        'Topic :: Other/Nonlisted Topic',
        'Topic :: Software Development :: Interpreters',
        'Operating System :: OS Independent',

        'License :: OSI Approved :: GNU Affero General Public License v3',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='asciidots esolang esoteric ascii dataflow language programming fun',

    packages=['dots'],
    scripts=['interpret.py'],

    package_data={'dots': glob.glob('dots/libs/*.dots')},

    data_files=[
        ('dots', glob.glob('dots/libs/*.dots'))
    ],

    install_requires=['Click'],

    entry_points='''
        [console_scripts]
        asciidots=interpret:main
    ''',
)



from dots.constants import DIRECTIONS, RIGHT, LEFT, UP, DOWN
from dots.exceptions import DotsExit
from dots.vector import Pos
from dots.chars import SingletonLibInnerWarpChar


def move_first_time(func):
    def _decorator(self, *args, **kwargs):
        # print('self is {0}'.format(self))
        # print(self.a)
        if not self.hasRun:
            self.move_parent()
            self.hasRun = True
        else:
            func(self, *args, **kwargs)

    return _decorator


def autodetect_next_state(dot, char):
    return TravelState(dot).next(char)


class State(object):
    def __init__(self, parent):
        """
        Describe the state of a dot with functions to get to the next state ofthe state machine.

        :param dots.dot.Dot parent:
        """

        self.env = parent.env
        self.parent = parent

        self.hasRun = False
        self.isWaiting = False
        self.id_mode = False

    def next(self, char):
        raise Exception("State Next Method Not Implemented!")

    def run(self, char):
        raise Exception("State Run Method Not Implemented!")

    def move_parent(self):
        self.parent.move()

    # define lambda like this: lambda dir: (your func or constant)
    def change_parent_dir_with_func(self, newDirLambda):
        self.parent.dir = newDirLambda(self.parent.dir)

    def set_parent_direction(self, direction):
        """
        Sets the direction of the dot

        :param dots.vector.Pos direction: The new direction of the dot.
        """

        self.parent.dir = direction

    def is_moving_vert(self):
        """True if the dot is moving verticaly."""
        return self.parent.dir.y != 0

    def is_moving_horiz(self):
        """True if the dot is moving horizontally."""
        return self.parent.dir.x != 0

    def __str__(self):
        return type(self).__name__

    def is_dead(self):
        return False

    def is_two_dots(self):
        return False


class TravelState(State):
    def next(self, char):
        if char == ' ':
            return DeadState(self.parent)
        elif char == '&' and not char.isOper():
            return ExitState(self.parent)
        elif char == '~':
            return TildeState(self.parent)
        elif char == '#':
            return ValueState(self.parent)
        elif char == '@':
            return IdState(self.parent)
        elif char == '$':
            return PrintState(self.parent)
        elif char in '[]{}' and self.is_moving_vert():
            return DeadState(self.parent)
        elif char.isSquareOper():
            return OperSquareState(self.parent)
        elif char.isCurlyOper():
            return OperCurlyState(self.parent)
        elif char == '-' and self.is_moving_vert():
            return DeadState(self.parent)
        elif char == '|' and self.is_moving_horiz():
            return DeadState(self.parent)
        elif char == ':' and self.parent.value == 0:
            return DeadState(self.parent)
        elif char == ';' and self.parent.value == 1:
            return DeadState(self.parent)
        else:
            return self

    def run(self, char):
        if char == '\\':
            self.change_parent_dir_with_func(lambda dir: Pos(dir.y, dir.x))
        elif char == '/':
            self.change_parent_dir_with_func(lambda dir: Pos(-dir.y, -dir.x))
        elif char == '(':
            self.set_parent_direction(RIGHT)
        elif char == ')':
            self.set_parent_direction(LEFT)
        elif char == '>':
            if self.is_moving_vert():
                self.set_parent_direction(RIGHT)
        elif char == '<':
            if self.is_moving_vert():
                self.set_parent_direction(LEFT)
        elif char == '^':
            if self.is_moving_horiz():
                self.set_parent_direction(UP)
        elif char == 'v':
            if self.is_moving_horiz():
                self.set_parent_direction(DOWN)
        elif char == '*':
            for dir in DIRECTIONS:

                if self.parent.dir in (dir, -dir):
                    continue

                next_pos = self.parent.pos + dir

                if self.env.world.does_loc_exist(next_pos) and self.env.world.get_char_at(next_pos) != ' ':

                    new_dot = self.parent.copy()
                    new_dot.dir = dir
                    new_dot.move()

                    self.env.dots.append(new_dot)
        elif isinstance(char, SingletonLibInnerWarpChar):
            self.parent.pos = self.parent.stack.pop()
        elif char.isWarp():
            if char.isSingletonLibWarp():
                self.parent.stack.append(self.parent.pos)

            self.parent.pos = char.get_dest_loc()
        else:
            pass

        self.move_parent()


class ValueState(State):
    def __init__(self, parent):
        super().__init__(parent)
        self.first_digit = True
        self.neg = False

    def next(self, char):
        if char.isdecimal() or char == '?':
            return self
        else:
            return autodetect_next_state(self.parent, char)

    @move_first_time
    def run(self, char):
        if char.isdecimal():
            if self.first_digit:
                self.parent.value = int(char)
                self.first_digit = False
            else:
                self.parent.value = self.parent.value * 10 + int(char)
        elif char == '?':
            try:
                self.parent.value = int(self.env.io.get_input())
            except ValueError:
                self.parent.value = 0

        self.move_parent()


class IdState(State):
    def __init__(self, parent):
        super().__init__(parent)
        self.first_digit = True
        self.setting_id = False

    def next(self, char):
        if char.isdecimal() or char == '?':
            return self
        elif self.setting_id:
            return autodetect_next_state(self.parent, char)
        elif char in '[]{}':
            if self.is_moving_vert():
                return DeadState(self.parent)
            else:
                return self
        elif char.isSquareOper():
            return OperSquareState(self.parent, id_mode=True)
        elif char.isCurlyOper():
            return OperCurlyState(self.parent, id_mode=True)
        elif char == '~':
            return TildeState(self.parent, id_mode=True)
        else:
            return autodetect_next_state(self.parent, char)

    @move_first_time
    def run(self, char):
        if char.isdecimal():
            self.setting_id = True
            if self.first_digit:
                self.parent.id = int(char)
                self.first_digit = False
            else:
                self.parent.id = self.parent.id * 10 + int(char)
        elif char == '?':
            try:
                self.parent.id = int(self.env.io.get_input())
            except ValueError:
                self.parent.id = 0
        else:
            pass

        self.move_parent()


class PrintState(State):
    def __init__(self, parent, newline=True):
        super().__init__(parent)
        self.newline = newline
        self.asciiMode = False
        self.pendingExit = False

    def next(self, char):
        if self.pendingExit:
            return autodetect_next_state(self.parent, char)
        elif char in '$_a#@':
            return self
        elif char == ' ':
            return DeadState(self.parent)
        elif char == '"':
            return PrintDoubleQuoteState(self.parent, newline=self.newline)
        elif char == "'":
            return PrintSingleQuoteState(self.parent, newline=self.newline)
        else:
            return autodetect_next_state(self.parent, char)

    @move_first_time
    def run(self, char):
        if char == '_':
            self.newline = False
        elif char == 'a':
            self.asciiMode = True
        elif char == '#':
            data = self.parent.value

            if self.asciiMode:
                data = chr(data)

            if self.newline:
                data = str(data) + '\n'

            self.env.io.on_output(data)
            self.pendingExit = True
        elif char == '@':
            data = self.parent.id

            if self.asciiMode:
                data = chr(data)

            if self.newline:
                data = str(data) + '\n'

            self.env.io.on_output(data)
            self.pendingExit = True
        else:
            pass

        self.move_parent()


class PrintDoubleQuoteState(PrintState):
    def __init__(self, parent, newline=True):
        super().__init__(parent, newline)
        self.text_buffer = ''

    def next(self, char):
        if self.pendingExit:
            return autodetect_next_state(self.parent, char)
        else:
            return self

    @move_first_time
    def run(self, char):
        if char == '"':
            if self.newline:
                self.text_buffer += '\n'

            self.env.io.on_output(self.text_buffer)

            self.pendingExit = True
        else:
            self.text_buffer += char

        self.move_parent()


class PrintSingleQuoteState(PrintState):

    def next(self, char):
        if self.pendingExit:
            return autodetect_next_state(self.parent, char)
        else:
            return self

    @move_first_time
    def run(self, char):
        if char == "'":
            if self.newline:
                self.env.io.on_output('\n')

            self.pendingExit = True
        else:
            self.env.io.on_output(char)

        self.move_parent()


class TwoDotState(State):
    def __init__(self, parent, isMaster, id_mode=None):
        super().__init__(parent)

        if id_mode is None:
            self.id_mode = False
        else:
            self.id_mode = id_mode

        self.isMaster = isMaster(self)

        self.isWaiting = True
        self.age = 0

    def is_two_dots(self):
        return True

    def next(self, char):
        if self.isWaiting:
            return self
        else:
            return autodetect_next_state(self.parent, char)

    def run(self, char):
        self.isWaiting = True

        if self.isMaster:
            # We want to find the companion dot that has been waiting the
            # longest. At the same time, if we find a dot that is also a master
            # that has been waiting longer than this, keep on waiting

            wait_here = [d for  d in self.env.dots if d.pos == self.parent.pos and d.state.is_two_dots()]
            masters_here = [d for d in wait_here if d.state.isMaster]
            other_here = [d for d in wait_here if not d.state.isMaster]

            if masters_here and other_here:
                oldest_master = max(masters_here, key=lambda d:d.state.age)
                oldest_other = max(other_here, key=lambda d:d.state.age)

                if oldest_master is self.parent:
                    candidate = oldest_other

                    if candidate.state.id_mode:
                        candidate_par = candidate.id
                    else:
                        candidate_par = candidate.value

                    if self.id_mode:
                        self_par = self.parent.id
                    else:
                        self_par = self.parent.value

                    self.do_operation(char, self_par, candidate_par, candidate)

                    candidate.state = DeadState(candidate)

                    # FIXME: This is not a good idea...
                    self.parent.state = TravelState(self.parent)
                    self.move_parent()

                    self.isWaiting = False

        self.age += 1

    def do_operation(self, *args, **kwargs):
        raise Exception("DoOperation not implemented!")


class OperState(TwoDotState):
    def __init__(self, parent, isMaster, id_mode=False):
        TwoDotState.__init__(self, parent, isMaster, id_mode)

    def do_operation(self, char, self_par, candidate_par, candidate):
        result = char.calc(self_par, candidate_par) * 1
        # NOTE: the `* 1` converts a boolean into an integer

        if self.id_mode:
            self.parent.id = result
        else:
            self.parent.value = result


class OperSquareState(OperState):
    def __init__(self, parent, id_mode=False):
        OperState.__init__(self, parent,
                           isMaster=(lambda self: self.is_moving_vert()),
                           id_mode=id_mode)


class OperCurlyState(OperState):
    def __init__(self, parent, id_mode=False):
        OperState.__init__(self, parent,
                           isMaster=(lambda self: self.is_moving_horiz()),
                           id_mode=id_mode)


class TildeState(TwoDotState):
    def __init__(self, parent, id_mode=False):
        TwoDotState.__init__(self, parent,
                             isMaster=(lambda self: self.is_moving_horiz()),
                             id_mode=id_mode)

    def do_operation(self, char, self_par, candidate_par, candidate):
        if candidate_par != 0:
            self.set_parent_direction(UP)


class DeadState(State):
    def next(self, char):
        return self

    def run(self, char):
        pass

    def is_dead(self):
        return True


class ExitState(State):
    def next(self, char):
        raise DotsExit

    def run(self, char):
        raise DotsExit



class Env(object):
    """
    A container for all the importants parts of an Asciidots program.
    """

    def __init__(self, world=None, dots=None, io=None, interpreter=None):
        """
        Structure for the for important parts of a dots environment.

        If you don't pass one of the 4 parameters, you need to set them quickly.

        :type world: dots.world.World
        :type dots: list[dots.dot.Dot]
        :type io: dots.callbacks.IOCallbacksStorage
        :type interpreter: dots.interpreter.AsciiDotsInterpreter
        """

        self.world = world
        self.dots = dots
        self.io = io
        self.interpreter = interpreter
